/* 
   Actor.cpp

   IceCreamDrop 
   Copyright (C) 2004 Marcus W Dillavou <line72@users.sf.net>
                  and Amy L Dillavou <gs2luv@users.sf.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include "Actor.h"

#include "World.h"

Actor::Actor()
  : Object()
{
  vector < string > animationNames;
  vector < int> animations;

  // still
  animationNames.push_back("mesh/Duke_Still.jba");
  animations.push_back(STILL);
  // walk
  animationNames.push_back("mesh/Duke_Walk.jba");
  animations.push_back(WALK);

  loadMesh("mesh/Duke.jbm", animationNames, animations);
 
  _iceCreamCone = new IceCreamCone();

  _collisionRadius = 1.0;

  // set the keys to a defaul
  _keyUp = SDLK_UP;
  _keyDown = SDLK_DOWN;
  _keyLeft = SDLK_LEFT;
  _keyRight = SDLK_RIGHT;
}

Actor::~Actor()
{
  if (_iceCreamCone)
    delete _iceCreamCone;
}

/**
 * make the character run
 */
void Actor::run()
{
}

/**
 * draw the actor and it's code
 */
void Actor::draw()
{
  glPushMatrix();
  
  glTranslatef(_currentTranslate.getX() + _initialTranslate.getX(),
	       _currentTranslate.getY() + _initialTranslate.getY(),
	       _currentTranslate.getZ() + _initialTranslate.getZ());
  glRotatef(_currentRotate.getX(), 1, 0, 0);
  glRotatef(_currentRotate.getY(), 0, 1, 0);
  glRotatef(_currentRotate.getZ(), 0, 0, 1);

  // scale up the character a little
  glPushMatrix();
  glScalef(1.5, 1.5, 1.5);
  _mesh->draw();

  jb::Matrix matrix;

  _mesh->GetTransform(19, matrix);

  glPushMatrix();
  glMultMatrixf((GLfloat *)&matrix);
  glScalef(2.0 / 3.0, 2.0 / 3.0, 2.0 / 3.0);
  // tranlsate slightly to the left
  glTranslatef(-0.2, -0.2, 0.0);
  _iceCreamCone->draw();
  glPopMatrix();

  glPopMatrix(); // pop scale

  glPopMatrix(); // pop all
}

/**
 * handle the input
 */
void Actor::handleInput()
{
  Uint8 *keyState = SDL_GetKeyState (NULL);
  double radians = 0.0;
  Camera *camera = Camera::getInstance();
  int numScoops = getCone()->getNumIceCreamScoops();
  double reduceSpeed = 0.0;
  if (numScoops == 0)
    reduceSpeed = 0.5;
  else
    reduceSpeed = (1.0 / numScoops) * 0.5;
  double speed = (camera->getSpeed() * reduceSpeed) * camera->getCurrentFPS();
  bool move = false;

  Vector3 rotate = getRotate();

  if (keyState[_keyUp] && keyState[_keyLeft]) // move up and left at 45
  {
    radians = 315 * (M_PI / 180.0);
    rotate.setY(45);
    move = true;
  }
  else if (keyState[_keyUp] && keyState[_keyRight]) // move up and right at 45
  {
    radians = 225 * (M_PI / 180.0);
    rotate.setY(315);
    move = true;
  }
  else if (keyState[_keyUp]) // move up
  {
    radians = 270 * (M_PI / 180.0);
    rotate.setY(0);
    move = true;
  }
  else if (keyState[_keyDown] && keyState[_keyLeft]) // down left at 45
  {
    radians = 45 * (M_PI / 180.0);
    rotate.setY(135);
    move = true;
  }
  else if (keyState[_keyDown] && keyState[_keyRight]) // down right at 45
  {
    radians = 135 * (M_PI / 180.0);
    rotate.setY(225);
    move = true;
  }
  else if (keyState[_keyDown]) // down
  {
    radians = 90 * (M_PI / 180.0);
    rotate.setY(180);
    move = true;
  }
  else if (keyState[_keyLeft]) // just left
  {
    radians = 0 * (M_PI / 180.0);
    rotate.setY(90);
    move = true;
  }
  else if (keyState[_keyRight]) // just right
  {
    radians = 180 * (M_PI / 180.0);
    rotate.setY(270);
    move = true;
  }
   
  if (move)
  {
    changeAction(WALK);
    
    // move the character
    double x, z;
    x = cos(radians) * speed;
    z = sin(radians) * speed;
    z = -z; // reverse z
    
    Vector3 v = getTranslate();
    v.setX(v.getX() + x);
    v.setZ(v.getZ() + z);
    // make sure this positions isn't outside the world !
    World *w = World::getInstance();
    
    // get the distance from the center
    double distance = sqrt((v.getX() * v.getX()) + (v.getZ() * v.getZ()));
    
    if (distance > -w->getCollisionRadius() &&
	distance < w->getCollisionRadius())
      setTranslate(v);
    
    setRotate(rotate);
  }
  else
    changeAction(STILL);
}

/**
 * see if we caught the ice cream
 */
void Actor::checkIceCreamCollision(vector < IceCreamScoop * > *scoops)
{
  vector < IceCreamScoop * > ::iterator it;
  //for (int i = 0; i < scoops.size(); i++)
  for (it = scoops->begin(); it != scoops->end(); it++)
  {
    // check x and z position
    Vector3 scoopI = (*it)->getInitialPosition();
    Vector3 scoopT = (*it)->getTranslate();
    if ((fabs(scoopI.getX() - (_initialTranslate.getX() + _currentTranslate.getX())) < _collisionRadius) &&
	(fabs(scoopI.getZ() - (_initialTranslate.getY() + _currentTranslate.getZ())) < _collisionRadius))
    {
      double height = (double)(getCone()->getNumIceCreamScoops()) * 0.5 + 1;

      if (fabs((scoopI.getY() + scoopT.getY()) - height) < _collisionRadius)
      {
	Vector3 reset;
	// reset the positions back to 0,0,0
	(*it)->setInitialPosition(reset);
	(*it)->setTranslate(reset);
	// we no longer want this to have a shadow
	(*it)->hasShadow(false);
	
	// add it to the cone
	getCone()->addIceCreamScoop(*it);

	// erase it from the vector
	scoops->erase(it);
	break;
      }
    }
  }
}

/**
 * set the keys, up down left and right
 */
void Actor::setKeys(int up, int down, int left, int right)
{
  _keyUp = up;
  _keyDown = down;
  _keyLeft = left;
  _keyRight = right;
}
