/*
  Animadead: A Skeletal Animation Library
  Copyright (C) 2004 John C. Butterfield

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  John C. Butterfield
  johnb003@hotmail.com
*/

#include "animadead.h"
#include "vertex.h"

#include <iostream>
#include <string>

#define ANIM_FPS 15

using namespace std;
using namespace jb;


/***************************************
			 Mesh Stuff
***************************************/

///////////////////
// SuperVertex
/////////////////
SuperVertex::SuperVertex() : weights(NULL)
{
}

SuperVertex::~SuperVertex()
{
  if (weights) delete [] weights;
}

///////////////////
// Face
/////////////////
Face::Face() : indices(NULL)
{
}

Face::~Face()
{
  if (indices) delete [] indices;
}

///////////////////
// Surface
/////////////////

Surface::Surface()
  : texturename(NULL), superVerts(NULL), faces(NULL)
{
}

void Surface::scaleto(const Surface &cp)
{
  numVerts = cp.numVerts;
  superVerts = new SuperVertex[numVerts];
  // SuperVertex has no relivant structure to scale to.
}

Surface::~Surface()
{
  if (texturename) delete [] texturename;
  if (superVerts) delete [] superVerts;
  if (faces) delete [] faces;
}

///////////////////
// Mesh
/////////////////
//string Mesh::overwriteTexturePath = "";

Mesh::Mesh(string filename, UserFunction *textureloader) : influences(NULL), surfaces(NULL)
{
  FILE *infile = fopen(filename.c_str(), "rb");
  if (!infile)
  {
    cout << "Could not open the file, " << filename << endl;
    throw(1);
  }

  unsigned int i,j,k;
  // the very first 2 bytes are used to determine endianness
  unsigned short test;
  fread(&test, sizeof(unsigned short), 1, infile);
  if (test != 0x00FF) {
    if (test == 0xFF00)
    {
      cout << "File format has wrong byte order. (I'll run a converter here later)" << endl;
      return;
    } else {
      cout << "Not jbm format" << endl;
      return;
    }
  }

  fread(&numInfluence, sizeof(unsigned int), 1, infile);
  if (numInfluence) influences = new unsigned int[numInfluence];
  for (i = 0; i < numInfluence; i++)
  {
    fread(&influences[i], sizeof(unsigned int), 1, infile);
  }

  fread(&numSurfaces, sizeof(unsigned int), 1, infile);
  surfaces = new Surface[numSurfaces];
  for (i = 0; i < numSurfaces; i++)
  {
    fread(&surfaces[i].nameLength, sizeof(unsigned int), 1, infile);
    surfaces[i].texturename = new char[surfaces[i].nameLength+1];
    fread(&surfaces[i].texturename[0], sizeof(char), surfaces[i].nameLength, infile);
    surfaces[i].texturename[surfaces[i].nameLength] = 0; // add null terminator
    if (textureloader)
      (*textureloader)(surfaces[i].texturename, surfaces[i].userData);

    fread(&surfaces[i].numVerts, sizeof(unsigned int), 1, infile);
    surfaces[i].superVerts = new SuperVertex[surfaces[i].numVerts];
    for (j = 0; j < surfaces[i].numVerts; j++)
    {
      fread(&surfaces[i].superVerts[j].x, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].y, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].z, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].a, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].b, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].c, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].u, sizeof(float), 1, infile);
      fread(&surfaces[i].superVerts[j].v, sizeof(float), 1, infile);
      if (numInfluence) surfaces[i].superVerts[j].weights = new float[numInfluence];
      for (k = 0; k < numInfluence; k++)
      {
	fread(&surfaces[i].superVerts[j].weights[k], sizeof(float), 1, infile);
	if (surfaces[i].superVerts[j].weights[k])
	  surfaces[i].superVerts[j].nonzeros.push_back(k);
      }
    }

    fread(&surfaces[i].numFaces, sizeof(unsigned int), 1, infile);
		
    surfaces[i].faces = new Face[surfaces[i].numFaces];
    for (j = 0; j < surfaces[i].numFaces; j++)
    {
      fread(&surfaces[i].faces[j].numIndices, sizeof(unsigned int), 1, infile);
      surfaces[i].faces[j].indices = new unsigned int[surfaces[i].faces[j].numIndices];
      for (unsigned int kk = 0; kk < surfaces[i].faces[j].numIndices; kk++)
      {
	fread(&surfaces[i].faces[j].indices[kk], sizeof(unsigned int), 1, infile);
      }
    }
  }
}

Mesh::Mesh() : influences(NULL), surfaces(NULL)
{
}

Mesh::~Mesh()
{
  if (influences) delete [] influences;
  if (surfaces) delete [] surfaces;
}

void Mesh::scaleto(const Mesh &cp)
{
  numSurfaces = cp.numSurfaces;
  surfaces = new Surface[numSurfaces];
  for (unsigned int i = 0; i < numSurfaces; i++)
  {
    surfaces[i].scaleto(cp.surfaces[i]);
  }
}

///////////////////
// MeshCacher
/////////////////
MeshCacher::MeshCacher() // private
{
}

MeshCacher::~MeshCacher()
{
  map<string, Mesh *>::iterator it;
  for (it = cache.begin(); it != cache.end(); it++)
  {
    delete it->second;
  }
}

static unsigned int meshInstCount = 0;
MeshCacher *MeshCacher::inst = 0; // static instanciated
MeshCacher *MeshCacher::instance()
{
  meshInstCount++;
  if (meshInstCount == 1) inst = new MeshCacher();
  return inst;
}

void MeshCacher::release()
{
  if (!inst) return;
  meshInstCount--;
  if (meshInstCount == 0) delete inst;
}

Mesh *MeshCacher::GetMesh(string filename, UserFunction *fp)
{
  map<string, Mesh *>::iterator it;
  if ((it = cache.find(filename)) != cache.end())
  {
    return it->second;
  }
  else
  {
    Mesh *temp = new Mesh(filename,fp);
    cache[filename] = temp;
    return temp;
  }
}

/***************************************
			 Anim Stuff
***************************************/

///////////////////
// Frame
/////////////////
Frame::Frame()
  : bones(0)
{
}

Frame::~Frame()
{
  if (bones) delete [] bones;
}

///////////////////
// Anim
/////////////////
Anim::Anim(string filename)
  : frames(0)
{
  FILE *infile = fopen(filename.c_str(), "rb");
  if (!infile)
  {
    cout << "Could not open the file, " << filename << endl;
    throw(1);
  }

  unsigned short test;
  fread(&test, sizeof(unsigned short), 1, infile);
  if (test != 0x00FE) {
    if (test == 0xFE00)
    {
      cout << "File format has wrong byte order. (I'll run a converter here later)" << endl;
      return;
    } else {
      cout << "Not jba format." << endl;
      return;
    }
  }

  fread(&numFrames, sizeof(unsigned int), 1, infile);

  unsigned int numBones;
  unsigned int i,j;
  if (numFrames)
  {
    fread(&numBones, sizeof(unsigned int), 1, infile);
    frames = new Frame[numFrames];
  }

  for (i = 0; i < numFrames; i++)
  {
    frames[i].bones = new Matrix[numBones];
    for (j = 0; j < numBones; j++)
      fread(frames[i].bones[j].matrix, sizeof(float), 16, infile);
    if (i != 0)
    {
      // the kfv for the previous frame = the kfv for this frame - the kfv for the previous frame
      frames[i-1].keyframeVelocity.x = frames[i].bones[0].matrix[3][0] - frames[i-1].bones[0].matrix[3][0];
      frames[i-1].keyframeVelocity.y = frames[i].bones[0].matrix[3][1] - frames[i-1].bones[0].matrix[3][1];
      frames[i-1].keyframeVelocity.z = frames[i].bones[0].matrix[3][2] - frames[i-1].bones[0].matrix[3][2];
    }
    if (i == numFrames-1) // last one, set the velocity for the last frame, which since it loops, should be the first frame anyway
    {
      if (i != 0) // last frame IS the first frame so the velocity for the first frame should work fine for the last
	frames[i].keyframeVelocity = frames[0].keyframeVelocity;
      else // there is only 1 frame (set velocity to zero)
	frames[i].keyframeVelocity.set(0,0,0);
    }
  }
  fclose(infile);

  for (i = 0; i < numFrames; i++)
  {
    // I don't want to overwrite the root bone until it's the last one
    // j is unsigned so we can't use -1 as < 0, it would wrap to maxint and be > 0 again
    for (j = numBones-1; j != (unsigned int)-1; j--)
    {
      frames[i].bones[j].matrix[3][0] -= frames[i].bones[0].matrix[3][0];
      frames[i].bones[j].matrix[3][1] -= frames[i].bones[0].matrix[3][1];
      frames[i].bones[j].matrix[3][2] -= frames[i].bones[0].matrix[3][2];
    }
  }
}

Anim::~Anim()
{
  if (frames) delete [] frames;
}

///////////////////
// AnimCacher
/////////////////
AnimCacher::AnimCacher()
{
}

AnimCacher::~AnimCacher()
{
  map<string, Anim *>::iterator it;
  for (it = cache.begin(); it != cache.end(); it++)
  {
    delete it->second;
  }
}

AnimCacher *AnimCacher::inst = 0;
static unsigned int animInstCount = 0;
AnimCacher *AnimCacher::instance()
{
  animInstCount++;
  if (animInstCount == 1) inst = new AnimCacher();
  return inst;
}

void AnimCacher::release()
{
  if (!inst) return;
  animInstCount--;
  if (animInstCount == 0) delete inst;
}

Anim *AnimCacher::GetAnim(string filename)
{
  map<string, Anim *>::iterator it;
  if ((it = cache.find(filename)) != cache.end())
  {
    return it->second;
  }
  else
  {
    Anim *temp = new Anim(filename);
    cache[filename] = temp;
    return temp;
  }
}

///////////////////
// AnimPlayer
/////////////////
AnimPlayer::AnimPlayer(Anim *anim, unsigned int startFrame) : anim(anim)
{
  previousFrame = startFrame;
  currentPosition = new Vertex();
  lastPosition = new Vertex();
  _animFinished = true;
}

AnimPlayer::~AnimPlayer()
{
  if (currentPosition) delete currentPosition;
  if (lastPosition) delete lastPosition;
}


void AnimPlayer::GetKeyframeVelocity(Vertex *tempVel)
{
  *tempVel = anim->frames[previousFrame].keyframeVelocity * ANIM_FPS; // returns meters per second
}

void AnimPlayer::GetVelocity(Vertex *tempVel)
{
  Vertex a(anim->frames[previousFrame].keyframeVelocity * (1-timeweight));
  Vertex b(anim->frames[previousFrame+1].keyframeVelocity * timeweight);
  *tempVel = (a + b) * ANIM_FPS; // returns meters per second
}

// this is only gets called when MovingOrigin(true)
void AnimPlayer::Update(float time)
{
  timeweight = time;
}

void AnimPlayer::Update(unsigned int framesPassed)
{
  previousFrame += framesPassed;
  if (previousFrame >= anim->numFrames - 1) // last indexable frame
  {
    previousFrame %= anim->numFrames-1;	// wrap back to the beginning
    _animFinished = true;
  }
  else
    _animFinished = false;
}

/***************************************
			AniModel Base
***************************************/
Model::Model(ModelType type, UserFunction *texture_loader)
  : originSubtraction(false), velocityControl(0), timeweight(0), fp(texture_loader), type(type)
{
  if (type == DEFORMABLE)
  {
    previousMesh = new vector<Mesh *>;
    nextMesh = new vector<Mesh *>;
  } else
  {
    previousMesh = NULL;
    nextMesh = NULL;
  }

  _currentAnim = 0;
}

Model::~Model()
{
  // delete meshes
  if (type == DEFORMABLE)
  {
    vector<Mesh *>::iterator it;
    for (it = previousMesh->begin(); it != previousMesh->end(); ++it)
    {
      delete *it;
    }
    for (it = nextMesh->begin(); it != nextMesh->end(); ++it)
    {
      delete *it;
    }
    delete previousMesh;
    delete nextMesh;
  }

  // bindMesh are deleted in MeshCacher
  // anims are deleted in AnimCacher
}

// when adding or swapping a mesh, we need to account for the next and prev meshes
// it would be best if you add or swap a mesh BEFORE you update, so we only have to
// make the next mesh accurate, and the update will take care of the previous mesh
unsigned int Model::AddMesh(string filename)
{
  bindMesh.push_back(MeshCacher::instance()->GetMesh(filename,fp));

  if (type == DEFORMABLE)
  {
    // we'll start by creating a previous and next mesh which consists of only data that changes during animation.
    // the scaleto function allocates the structures to match the size of the bind Mesh where relivant.
    Mesh *temp = new Mesh();
    temp->scaleto(*bindMesh.back());
    nextMesh->push_back(temp);

    temp = new Mesh();
    temp->scaleto(*bindMesh.back());
    previousMesh->push_back(temp);

    // now we'll update our nextmesh
    // OPTIMIZE: this really only needs to update the mesh that's being added (the others should still be valid)
    UpdateMeshes();
  }

  return (unsigned int)bindMesh.size()-1; 
}

// well, our copied meshes for next and previous are no longer the valid size
// it wouldn't be very feasible to reuse them
// (even though we could if the previous one is larger in all dimensions)
void Model::SwapMesh(unsigned int index, string filename)
{
  bindMesh[index] = MeshCacher::instance()->GetMesh(filename,fp);
	
  if (type == DEFORMABLE)
  {
    // these are no longer the valid size
    delete (*nextMesh)[index];
    delete (*previousMesh)[index];

    // so we'll make the valid sized ones
    (*nextMesh)[index]=new Mesh();
    (*nextMesh)[index]->scaleto(*bindMesh[index]);
	
    (*previousMesh)[index]=new Mesh();
    (*previousMesh)[index]->scaleto(*bindMesh[index]);

    // and update the nextmesh
    // (again previous mesh should be taken care of by the real update call)
    // OPTIMIZE: this really only needs to update the mesh that's being added (the others should still be valid)
    UpdateMeshes();
  }
}

unsigned int Model::AddAnim(std::string filename, unsigned int startFrame)
{
  AnimPlayer *anim = new AnimPlayer(AnimCacher::instance()->GetAnim(filename), startFrame);
  anims.push_back(anim);
  return (unsigned int)anims.size()-1;
}

void Model::SwapAnim(unsigned int index, std::string filename, unsigned int startFrame)
{
  delete anims[index];
  anims[index] = new AnimPlayer(AnimCacher::instance()->GetAnim(filename), startFrame);
}

void Model::Bind(unsigned int bone, unsigned int anim)
{
  // resize the vector to account for more bones
  // I keep telling myself this is a bad solution, but I'm forgetting why :P
  if (bone >= animFromBone.size())
    animFromBone.resize(bone+1);
  animFromBone[bone] = anim;
  
  _currentAnim = anims[anim];
}

void Model::GetTransform(unsigned int bone, Matrix &matrix)
{
  AnimPlayer *animp = anims[animFromBone[bone]];
  Matrix *matrix1 = &animp->anim->frames[animp->previousFrame].bones[bone];
  Matrix *matrix2 = &animp->anim->frames[(animp->previousFrame+1 != animp->anim->numFrames)?animp->previousFrame+1:0].bones[bone];
  matrix.matrix[0][0] = matrix1->matrix[0][0]*(1-timeweight) + matrix2->matrix[0][0]*timeweight;
  matrix.matrix[1][0] = matrix1->matrix[1][0]*(1-timeweight) + matrix2->matrix[1][0]*timeweight;
  matrix.matrix[2][0] = matrix1->matrix[2][0]*(1-timeweight) + matrix2->matrix[2][0]*timeweight;
  matrix.matrix[3][0] = matrix1->matrix[3][0]*(1-timeweight) + matrix2->matrix[3][0]*timeweight;
  matrix.matrix[0][1] = matrix1->matrix[0][1]*(1-timeweight) + matrix2->matrix[0][1]*timeweight;
  matrix.matrix[1][1] = matrix1->matrix[1][1]*(1-timeweight) + matrix2->matrix[1][1]*timeweight;
  matrix.matrix[2][1] = matrix1->matrix[2][1]*(1-timeweight) + matrix2->matrix[2][1]*timeweight;
  matrix.matrix[3][1] = matrix1->matrix[3][1]*(1-timeweight) + matrix2->matrix[3][1]*timeweight;
  matrix.matrix[0][2] = matrix1->matrix[0][2]*(1-timeweight) + matrix2->matrix[0][2]*timeweight;
  matrix.matrix[1][2] = matrix1->matrix[1][2]*(1-timeweight) + matrix2->matrix[1][2]*timeweight;
  matrix.matrix[2][2] = matrix1->matrix[2][2]*(1-timeweight) + matrix2->matrix[2][2]*timeweight;
  matrix.matrix[3][2] = matrix1->matrix[3][2]*(1-timeweight) + matrix2->matrix[3][2]*timeweight;
  matrix.matrix[0][3] = 0;
  matrix.matrix[1][3] = 0;
  matrix.matrix[2][3] = 0;
  matrix.matrix[3][3] = 1;
}

void Model::MovingOrigin(bool value)
{
  originSubtraction = value;
}

void Model::OriginAnim(unsigned int index)
{
  velocityControl = anims[index];
}

void Model::GetKeyframeVelocity(Vertex *velocity)
{
  velocityControl->GetKeyframeVelocity(velocity);
}

void Model::GetVelocity(Vertex *velocity)
{
  velocityControl->GetVelocity(velocity);
}

void Model::Update(float timestep)
{
  // shouldn't be called on Static models
  timeweight += timestep * 15; // 15 fps

  vector<AnimPlayer *>::iterator it;
  for (it = anims.begin(); it != anims.end(); ++it)
  {
    (*it)->Update(timeweight - (unsigned int)timeweight);
  }

  if (timeweight >= 1)
  {
    // every keyframe
    vector<AnimPlayer *>::iterator it;
    for (it = anims.begin(); it != anims.end(); ++it)
    {
      (*it)->Update((unsigned int)timeweight);
    }

    timeweight -= (unsigned int)timeweight;
    UpdateMeshes();
  }
}

void Model::UpdateMeshes()
{
  // I am not going to check for animations on each bone.
  // if this is called it's going to have animations loaded.

  // loop through the components of each mesh at the previous frame
  // go through the bones they are attached to, and find what anim that bone is associated with
  // use the current orientation of the bone in that anim's prev and next frame

  vector<Mesh *> *temp = previousMesh;
  previousMesh = nextMesh;
  nextMesh = temp;
  for (unsigned int mesh = 0; mesh < bindMesh.size(); mesh++)
  {
    for (unsigned int surf = 0; surf < bindMesh[mesh]->numSurfaces; surf++)
    {
      for (unsigned int vert = 0; vert < bindMesh[mesh]->surfaces[surf].numVerts; vert++)
      {
	SuperVertex *vertex = &(*nextMesh)[mesh]->surfaces[surf].superVerts[vert];
	SuperVertex *bindvertex = &bindMesh[mesh]->surfaces[surf].superVerts[vert];

	float bindx = bindvertex->x;
	float bindy = bindvertex->y;
	float bindz = bindvertex->z;
	float binda = bindvertex->a;
	float bindb = bindvertex->b;
	float bindc = bindvertex->c;

	vertex->x = 0;
	vertex->y = 0;
	vertex->z = 0;
	vertex->a = 0;
	vertex->b = 0;
	vertex->c = 0;

	vector<unsigned int>::iterator it;
	// this gives us the relivant bones for this vertex
	for (it = bindvertex->nonzeros.begin(); it != bindvertex->nonzeros.end(); ++it)
	{
	  // weight: bindvertex->weights[*it]
	  // bone: bindMesh[mesh]->influences[*it]
	  // animplayer: anims[animFromBone[bone]];
	  // next: animplayer->previousFrame + 1
	  // matrix: animplayer->anim->frame[next].bones[bone]
	  unsigned int bone = bindMesh[mesh]->influences[*it];
	  AnimPlayer *animp = anims[animFromBone[bone]];
	  Matrix *transform = &animp->anim->frames[animp->previousFrame+1].bones[bone]; // 
	  vertex->x += (	transform->matrix[0][0]*bindx +
				transform->matrix[1][0]*bindy +
				transform->matrix[2][0]*bindz +
				transform->matrix[3][0] ) * bindvertex->weights[*it];

	  vertex->y += (	transform->matrix[0][1]*bindx +
				transform->matrix[1][1]*bindy +
				transform->matrix[2][1]*bindz +
				transform->matrix[3][1] ) * bindvertex->weights[*it];

	  vertex->z += (	transform->matrix[0][2]*bindx +
				transform->matrix[1][2]*bindy +
				transform->matrix[2][2]*bindz +
				transform->matrix[3][2] ) * bindvertex->weights[*it];

	  vertex->a += (	transform->matrix[0][0]*binda +
				transform->matrix[1][0]*bindb +
				transform->matrix[2][0]*bindc ) * bindvertex->weights[*it];

	  vertex->b += (	transform->matrix[0][1]*binda +
				transform->matrix[1][1]*bindb +
				transform->matrix[2][1]*bindc ) * bindvertex->weights[*it];

	  vertex->c += (	transform->matrix[0][2]*binda +
				transform->matrix[1][2]*bindb +
				transform->matrix[2][2]*bindc ) * bindvertex->weights[*it];
	}
      }
    }
  }
}

