/* 
   Camera.cpp

   IceCreamDrop 
   Copyright (C) 2004 Marcus W Dillavou <line72@users.sf.net>
                  and Amy L Dillavou <gs2luv@users.sf.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include "Camera.h"

Camera::Camera ()
{
  // initialize everything
  init ();
}

Camera::~Camera ()
{
  // TODO: put destructor code here
}

/**
 * Initialize our variables
 */
void Camera::init ()
{
  _position.set (0.0, 0.0, 0.0); // start at the origin
  _view.set (0.0, 1.0, 0.5);
  _upVector.set (0.0, 0.0, 1.0);

  _fps = 0.0;
  _screenWidth = 1024;
  _screenHeight = 768;
  
  _speed = 25.0; // random initial speed
}

/**
 * Update the camera's data
 */
void Camera::update ()
{
  // find the cross product for the strafe vector
  Vector3 cross = Cross (_view - _position, _upVector);
  // Normalize it
  _strafe = Normalize (cross);

  cross = Cross(_view - _position, _upVector);
  _walk = Normalize(cross);
  
   // check for movment
  //checkForMovement ();

  // set the view from the mouse
  //setViewFromMouse ();
  
  // calculate the frame rate
  _fps = CalculateFrameRate ();
}

/**
 * Check for movement based on the keyboard input
 */
void Camera::checkForMovement ()
{
  double speed;
  speed = _speed * _fps;
  
  Uint8 *keyState = SDL_GetKeyState (NULL);
  
  if (keyState[SDLK_UP] || keyState[SDLK_w])	// up
    walk(-speed);
  if (keyState[SDLK_DOWN] || keyState[SDLK_s])	// down
    walk(speed);
  if (keyState[SDLK_LEFT] || keyState[SDLK_a])	// left
    strafeCamera (-speed);
  if (keyState[SDLK_RIGHT] || keyState[SDLK_d])	// right
    strafeCamera (speed);
}

/**
 * Changes the position, view, and upVector of the camera
 */
void Camera::positionCamera (double position[], double view[], double upVector[])
{
  _position.set (position);
  _view.set (view);
  _upVector.set (upVector);
}

void Camera::positionCamera (Vector3 position, Vector3 view, Vector3 upVector)
{
  _position.set (position.getVector ());
  _view.set (view.getVector ());
  _upVector.set (upVector.getVector ());
}

void Camera::positionCamera (double p1, double p2, double p3, double v1, double v2, double v3, double u1, double u2, double u3)
{
  double tmp[3];
  
  tmp[0] = p1; tmp[1] = p2; tmp[2] = p3;
  _position.set(tmp);
  tmp[0] = v1; tmp[1] = v2; tmp[2] = v3;
  _view.set(tmp);
  tmp[0] = u1; tmp[1] = u2; tmp[2] = u3;
  _upVector.set(tmp);
}

/** 
 * rotate the view around a position
 */
void Camera::rotateView (double angle, double x, double y, double z)
{
  Vector3 view1, view2;
  double cosTheta, sinTheta;
  
  view2 = _view - _position;
  
  // calculate the sine and cosine of the angle
  cosTheta = cos (angle);
  sinTheta = sin (angle);
  
  // find the new x position
  view1.setX ((cosTheta + (1 - cosTheta) * x * x) * view2.getX ());
  view1.setX (view1.getX () + ((1 - cosTheta) * x * y - z * sinTheta) * view2.getY ());
  view1.setX (view1.getX () + ((1 - cosTheta) * x * z + y * sinTheta) * view2.getZ ());
  
  // find the new y position
  view1.setY (((1 - cosTheta) * x * y + z * sinTheta) * view2.getX ());
  view1.setY (view1.getY () + (cosTheta + (1 - cosTheta) * y * y) * view2.getY ());
  view1.setY (view1.getY () + ((1 - cosTheta) * y * z - x * sinTheta) * view2.getZ ());
  
  // find the new z position
  view1.setZ (((1 - cosTheta) * x * z - y * sinTheta) * view2.getX ());
  view1.setZ (view1.getZ () + ((1 - cosTheta) * y * z + x * sinTheta) * view2.getY ());
  view1.setZ (view1.getZ () + (cosTheta + (1 - cosTheta) * z * z) * view2.getZ ());
  
  // now add the new vector to our position
  _view = _position + view1;
}

/**
 * move the camera based on the mouse
 */
void Camera::setViewFromMouse ()
{
  int middleX, middleY;
  int mouseX, mouseY;
  double angleY, angleZ;
  
  middleX = _screenWidth >> 1;
  middleY = _screenHeight >> 1;
  
  angleY = angleZ = 0.0;
  
  // get the cursors position
  SDL_GetMouseState (&mouseX, &mouseY);
  
  // if the cursor is in the middle, don't bother
  if ((mouseX == middleX) && (mouseY == middleY))
    return;
  
  // Set the mouse position to the middle of our window
  SDL_WarpMouse (middleX, middleY);
  
  // get the direction the mouse moved in
  angleY = (middleX - mouseX) / 500.0;	// divide by 500 so we don't move so fast
  angleZ = (middleY - mouseY) / 500.0;	// divide by 500 so we don't move so fast
  
  // keep track of the current rotation
  _currentRotationX -= angleZ;
  
  // cap the extremes
  if (_currentRotationX > 10.5)
    _currentRotationX = 10.5;
  else if (_currentRotationX < -10.5)
    _currentRotationX = -10.5;
  else
  {
    // get the perpendicular vector
    Vector3 axis = Cross (_view - _position, _upVector);
    axis = Normalize (axis);
    
    // rotate around our perpendicular axis and along the y axis
    rotateView (angleZ, axis.getX (), axis.getY (), axis.getZ ());
    rotateView (angleY, 0, 1, 0);
  }
}

/**
 * moves the camera forward and backward
 */
void Camera::moveCamera (double speed)
{
  // get the current view vector
  Vector3 view = _view - _position;
  // normalize it
  view = Normalize (view);
  
  _position.setX (_position.getX () + (view.getX () * speed));
  _position.setY (_position.getY () + (view.getY () * speed));
  _position.setZ (_position.getZ () + (view.getZ () * speed));
  _view.setX (_view.getX () + (view.getX () * speed));
  _view.setY (_view.getY () + (view.getY () * speed));
  _view.setZ (_view.getZ () + (view.getZ () * speed));
}

/**
 * strafes the camera left and right
 */
void Camera::strafeCamera(double speed)
{
  // add the strafe vector to our position
  _position.setX (_position.getX () + _strafe.getX () * speed);
  _position.setZ (_position.getZ () + _strafe.getZ () * speed);
  
  // add the strafe vector to our view
  _view.setX (_view.getX () + _strafe.getX () * speed);
  _view.setZ (_view.getZ () + _strafe.getZ () * speed);
}

/**
 * walks the camera forward and backward
 */
void Camera::walk(double speed)
{
  // add the strafe vector to our position
  //_position.setX (_position.getX () + _walk.getZ() * speed);
  _position.setZ (_position.getZ () + speed);
  
  // add the strafe vector to our view
  //_view.setX (_view.getX () + _walk.getX() * speed);
  _view.setZ (_view.getZ () + speed);

  /************
  // get the current view vector
  Vector3 view = _view - _position;
  // normalize it
  view = Normalize (view);
  
  _position.setX (_position.getX () + (view.getX () * speed));
  //_position.setY (_position.getY () + (view.getY () * speed));
  _position.setZ (_position.getZ () + (view.getZ () * speed));
  _view.setX (_view.getX () + (view.getX () * speed));
  //_view.setY (_view.getY () + (view.getY () * speed));
  _view.setZ (_view.getZ () + (view.getZ () * speed));
  ***********/
}

/**
 * Call glLookAt
 */
void Camera::look()
{
  //cout << _position.getX () << ", " << _position.getY () << ", " << _position.getZ () << ", " << _view.getX () << ", " << _view.getY () << ", " << _view.getZ () << ", " << _upVector.getX () << ", " << _upVector.getY () << ", " << _upVector.getZ () << endl;
  gluLookAt (_position.getX (), _position.getY (), _position.getZ (), _view.getX (), _view.getY (), _view.getZ (), _upVector.getX (), _upVector.getY (), _upVector.getZ ());
  //cout << _position << endl;
}

Camera *Camera::getInstance()
{
  static Camera c;
  
  return &c;
}
