/* 
   World.cpp

   IceCreamDrop 
   Copyright (C) 2004 Marcus W Dillavou <line72@users.sf.net>
                  and Amy L Dillavou <gs2luv@users.sf.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/


#include "World.h"

World::World()
  : Object()
{
  _name = "World";

  // load in the mesh
  loadMesh("mesh/IceCreamWorld.jbm");
  // set the radius of the world
  //  !mwd - this really shouldn't be hard coded !
  _collisionRadius = 6.5;

  // create two new characters and give them an
  //  initial position
  _player1 = new Actor();
  _player1->setInitialPosition(Vector3(-1.0, 0.0, 0.0));
  _player2 = new Actor();
  _player2->setInitialPosition(Vector3(1.0, 0.0, 0.0));
  // set the keys of player 2
  _player2->setKeys(SDLK_w, SDLK_s, SDLK_a, SDLK_d);

  // calcuate the falling height of the ice cream
  calculateMinMaxIceCreamHeight();

  // start off creating 3 scoops every 3 seconds
  _iceCreamScoopSeconds = 3;

  // create some ice cream scoops
  createNewScoops();

  // reset the scoop timer
  _scoopTimer.reset();
  // reset the game timer
  _gameTimer.reset();  
}

World::~World()
{
  if (_player1)
    delete _player1;
  if (_player2)
    delete _player2;
  for (int i = 0; i < _iceCreamScoops.size(); i++)
    delete _iceCreamScoops[i];
  _iceCreamScoops.clear();
}

/**
 * get an instance of the world
 */
World *World::getInstance()
{
  static World instance;

  return &instance;
}

/**
 * draw the world
 *  and all its objects
 */
void World::draw()
{
  // draw the world
  glDisable(GL_CULL_FACE);
  _mesh->draw();
  glEnable(GL_CULL_FACE);

  // draw the players
  _player1->draw();
  _player2->draw();

  // draw the ice cream scoops
  for (int i = 0; i < _iceCreamScoops.size(); i++)
    _iceCreamScoops[i]->draw();
}

/**
 * update the world
 *  and all its objects
 *
 * returns 
 * -1 if the game is still going on,
 * 0 if draw
 * 1 if player one won
 * 2 if player two won
 */
int World::update()
{
  if (_gameTimer.update() > 60) // each match last 60 seconds
  {
    if (_player1->getCone()->getNumIceCreamScoops() == _player2->getCone()->getNumIceCreamScoops()) // draw
      return 0;
    else if (_player1->getCone()->getNumIceCreamScoops() > _player2->getCone()->getNumIceCreamScoops()) // player 1 won
      return 1;
    else // player 2 won
      return 2;
  }

  // tell the ice cream scoops to start falling
  for (int i = 0; i < _iceCreamScoops.size(); i++)
    _iceCreamScoops[i]->fall();

  // create 3 new scoops if time is up
  if (_scoopTimer.update() > _iceCreamScoopSeconds)
  {
    createNewScoops();
    _scoopTimer.reset();
  }

  // check for collision
  checkCollision();

  // update the players movement
  updatePlayersMovement();

  // return -1 to contine with the game
  return -1;
}

/**
 * calculate the min and max ice cream height
 */
void World::calculateMinMaxIceCreamHeight()
{
  _minIceCreamHeight = 10;
  _maxIceCreamHeight = 15;
}

/**
 * create 3 new scoops at random positions
 */
void World::createNewScoops()
{
  // create 3 new scoops
  for (int i = 0; i < 3; i++)
  {
    IceCreamScoop *i = new IceCreamScoop();
    double x = 0, y = 0, z = 0;
    // get a random x, and z position inside the world radius
    x = (rand() % (int)(_collisionRadius * 2)) - _collisionRadius;
    z = (rand() % (int)(_collisionRadius * 2)) - _collisionRadius;
    // get a random y position high enough to be off screen, but not too high
    y = (rand() % (_maxIceCreamHeight - _minIceCreamHeight)) + _minIceCreamHeight;
    i->setInitialPosition(Vector3(x, y, z));
    _iceCreamScoops.push_back(i);
  }
}

/**
 * check to see if a player
 *  caught an ice cream or
 *  if it hit the floor
 */
void World::checkCollision()
{
  vector < IceCreamScoop * >::iterator it = _iceCreamScoops.begin();
  for (it = _iceCreamScoops.begin(); it != _iceCreamScoops.end(); it++)
  {
    //if ((*it)->getInitialPosition().getY() + (*it)->getTranslate().getY() < 0.5)
    if ((*it)->isDone())
    {
      delete (*it);
      _iceCreamScoops.erase(it);
    }
  }

  // make sure the two players haven't collided

  // check to see if the player caught the ice cream
  _player1->checkIceCreamCollision(&_iceCreamScoops);
  _player2->checkIceCreamCollision(&_iceCreamScoops);
}

/**
 * check for keyboard/joystick
 *  input for moving the player
 */
void World::updatePlayersMovement()
{
  // check for key presses
  _player1->handleInput();
  _player2->handleInput();

  // make sure we update the animations !
  _player1->updateAnimation();
  _player2->updateAnimation();
}
